# -*- coding: utf-8 -*-
"""
Utils for CLI.
"""

from __future__ import absolute_import, unicode_literals

import datetime
import io
import json
import logging
import os

from box import Box

from cfme_testcases import consts
from cfme_testcases.exceptions import TestcasesException
from dump2polarion import utils as d2p_utils

# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


class XMLsContainer(object):
    """Container for generated XMLs."""

    __slots__ = tuple(consts.OUT_FILES) + ("imported", "processed")

    def __init__(self):
        for prop in consts.OUT_FILES:
            setattr(self, prop, None)
        self.imported = set()
        self.processed = set()


def get_submit_args(args):
    """Gets arguments for the `submit_and_verify` method."""
    submit_args = dict(
        testrun_id=args.testrun_id,
        user=args.user,
        password=args.password,
        no_verify=args.no_verify,
        verify_timeout=args.verify_timeout,
        dry_run=args.dry_run,
    )
    submit_args = {k: v for k, v in submit_args.items() if v is not None}
    return Box(submit_args, frozen_box=True, default_box=True)


def get_job_logname(file_type, output_dir):
    """Returns filename for saving the log file."""
    return get_filename(file_type, output_dir, "job", "log")


def get_import_file_name(file_type, output_dir):
    """Generates filename for saving XML output."""
    return get_filename(file_type, output_dir, "import", "xml")


def get_filename(file_type, output_dir, prefix, ext):
    """Generates filename for file saving."""
    rec = consts.OUT_FILES.get(file_type)
    if not rec:
        raise TestcasesException("Unknown file type `{}`".format(file_type))
    key, file_name = rec

    key = "-{}".format(key) if key else ""
    file_name = "-{}".format(file_name) if file_name else ""
    stamp = "{:%Y%m%d%H%M%S}".format(datetime.datetime.now())

    return os.path.join(output_dir, "{}-{}{}{}.{}".format(prefix, stamp, key, file_name, ext))


def save_xmls(output_dir, xmls_container):
    """Saves the generated XML files if instructed to do so."""
    output_dir = output_dir or os.curdir

    for file_type in consts.OUT_FILES:
        xml_file = getattr(xmls_container, file_type, None)
        if xml_file is not None:
            d2p_utils.write_xml_root(xml_file, get_import_file_name(file_type, output_dir))


def check_lookup_methods(config):
    """Checks that lookup methods are configured correctly.

    Returns lookup method.
    """
    xunit_props = config.get("xunit_import_properties") or {}
    xunit_lookup = xunit_props.get("polarion-lookup-method") or "id"

    testcase_props = config.get("testcase_import_properties") or {}
    testcase_lookup = testcase_props.get("lookup-method") or "id"

    if xunit_lookup != testcase_lookup:
        raise TestcasesException("The lookup-method for test cases and XUnit must be the same.")

    if xunit_lookup != "custom":
        return xunit_lookup

    xunit_custom = xunit_props.get("polarion-custom-lookup-method-field-id")
    testcase_custom = testcase_props.get("polarion-custom-lookup-method-field-id")
    if xunit_custom != testcase_custom:
        raise TestcasesException(
            "The polarion-custom-lookup-method-field-id for test cases and "
            "XUnit must be the same."
        )

    return xunit_lookup


def load_json_file(json_file):
    """Returns content of JSON file."""
    with io.open(json_file, encoding="utf-8") as input_json:
        return json.load(input_json)



# -*- coding: utf-8 -*-
"""
Submits collected XML files to appropriate Importers in appropriate order.
"""

from __future__ import absolute_import, unicode_literals

import logging
import threading

from cfme_testcases import utils
from cfme_testcases.exceptions import NothingToDoException, TestcasesException
from dump2polarion import properties
from dump2polarion import submit as d2p_submit
from dump2polarion import utils as d2p_utils

# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


class CollectedDataSubmitter(object):
    """Submitter of collected XMLs."""

    def __init__(self, submit_args, config, xmls_container, output_dir=None):
        self.submit_args = submit_args
        self.config = config
        self.xmls_container = xmls_container
        self.output_dir = output_dir

    @staticmethod
    def get_job_log(file_type, output_dir):
        """Returns filename for saving the log file."""
        if not output_dir:
            return None
        return utils.get_job_logname(file_type, output_dir)

    @staticmethod
    def _append_msg(retval, msg, succeeded, failed):
        if retval:
            succeeded.append(msg)
        else:
            failed.append(msg)

    @staticmethod
    def _log_outcome(succeeded, failed):
        if succeeded and failed:
            logger.info("SUCCEEDED to %s.", ", ".join(succeeded))
        if failed:
            raise TestcasesException("FAILED to {}.".format(", ".join(failed)))

        logger.info("DONE - RECORDS SUCCESSFULLY UPDATED!")

    def _submit_xml(self, file_type):
        xml_root = getattr(self.xmls_container, file_type)

        if xml_root is None or file_type in self.xmls_container.processed:
            raise NothingToDoException("No data to import.")
        self.xmls_container.processed.add(file_type)

        job_log = self.get_job_log(file_type, self.output_dir)
        properties.remove_response_property(xml_root)
        response = d2p_submit.submit_and_verify(
            xml_root=xml_root, config=self.config, log_file=job_log, **self.submit_args
        )
        if response:
            self.xmls_container.imported.add(file_type)
        return response

    def update_testcases_threaded(self):
        """Updates existing testcases in new thread."""
        if self.xmls_container.updated_testcases is None:
            raise NothingToDoException("No data to import.")

        job_log = self.get_job_log("updated_testcases", self.output_dir)
        properties.remove_response_property(self.xmls_container.updated_testcases)
        all_submit_args = dict(
            xml_root=self.xmls_container.updated_testcases,
            config=self.config,
            log_file=job_log,
            **self.submit_args
        )

        # run it in separate thread so we can continue without waiting
        # for the submit to finish
        def _run_submit(results, args_dict):
            retval = d2p_submit.submit_and_verify(**args_dict)
            results.append(retval)

        output = []
        updating_testcases_t = threading.Thread(target=_run_submit, args=(output, all_submit_args))
        updating_testcases_t.start()

        return updating_testcases_t, output

    def submit_file_type(self, file_type, msg, succeeded, failed):
        """Submits specified file type."""
        imported = False
        try:
            imported = self._submit_xml(file_type)
            self._append_msg(imported, msg, succeeded, failed)
        except NothingToDoException:
            pass
        return imported

    def update_testrun(self, file_type, msg, succeeded, failed):
        """Updates testrun using XML file specified by file type."""
        testrun_id = self.submit_args.get("testrun_id") or d2p_utils.get_testrun_id_config(
            self.config
        )

        xml_root = getattr(self.xmls_container, file_type)
        imported = False
        if testrun_id:
            imported = self.submit_file_type(file_type, msg, succeeded, failed)
        elif not testrun_id and xml_root is not None:
            logger.warning("Not updating testrun, testrun ID not specified.")
        return imported

    def submit_all(self):
        """Submits all outstanding XMLs to Polarion Importers.

        There should be only XMLs meant for submit in the `self.xmls_container`.
        """
        if not self.submit_args:
            return

        succeeded, failed = [], []

        # requirements are needed by testcases, update them first
        self.submit_file_type("requirements", "update requirements", succeeded, failed)

        # update existing testcases in new thread
        updating_testcases_t = None
        try:
            updating_testcases_t, output = self.update_testcases_threaded()
        except NothingToDoException:
            pass

        # import missing data
        self.submit_file_type("missing_testcases", "add missing testcases", succeeded, failed)
        if (
            "missing_testcases" in self.xmls_container.imported
            or "testcases" in self.xmls_container.imported
        ):
            self.update_testrun("missing_testsuites", "update testrun", succeeded, failed)

        # wait for update of existing testcases to finish
        if updating_testcases_t:
            updating_testcases_t.join()
            response = output.pop()
            self._append_msg(response, "update existing testcases", succeeded, failed)
            if response:
                self.xmls_container.imported.add("updated_testcases")

        # import complete data
        self.submit_file_type("testcases", "add all testcases", succeeded, failed)
        if "testcases" in self.xmls_container.imported:
            self.update_testrun("testsuites", "create testrun", succeeded, failed)

        self._log_outcome(succeeded, failed)


def submit_all(submit_args, config, xmls_container, output_dir):
    """Submits collected XML files to Polarion Importers."""
    return CollectedDataSubmitter(submit_args, config, xmls_container, output_dir).submit_all()


def _do_import(submit_args, config, xmls_container, output_dir, file_type):
    xml_root = getattr(xmls_container, file_type)

    if xml_root is None or file_type in xmls_container.processed:
        return None
    xmls_container.processed.add(file_type)

    job_log = CollectedDataSubmitter.get_job_log(file_type, output_dir)
    properties.remove_response_property(xml_root)

    response = d2p_submit.submit_and_verify(
        xml_root=xml_root, config=config, log_file=job_log, **submit_args
    )

    if not response:
        raise TestcasesException("Failed to do import.")

    xmls_container.imported.add(file_type)

    return job_log


def import_testcases(submit_args, config, xmls_container, output_dir):
    """Imports unfiltered testcases into Polarion."""
    job_log = None
    try:
        job_log = _do_import(submit_args, config, xmls_container, output_dir, "testcases")
    except TestcasesException:
        raise TestcasesException("Failed to update existing testcases.")

    logger.info("SUCCEEDED to import testcases.")
    return job_log


def initial_submit(submit_args, config, xmls_container, output_dir):
    """Submits data to Polarion to get a list of missing testcases."""
    job_log = None
    try:
        job_log = _do_import(submit_args, config, xmls_container, output_dir, "init_testsuites")
    except TestcasesException:
        raise TestcasesException("Failed to do the initial submit.")

    logger.info("SUCCEEDED to do initial import.")
    return job_log



# -*- coding: utf-8 -*-
"""
Constants.
"""

from __future__ import unicode_literals

TESTS_DATA_JSON = "tests_data.json"

XUNIT = "xunit"
TEST_CASE = "testcases"
REQUIREMENTS = "requirements"

# file_type: (key, file_name)
OUT_FILES = {
    "testcases": (None, TEST_CASE),
    "testsuites": (None, XUNIT),
    "requirements": (None, REQUIREMENTS),
    "init_testsuites": ("init", XUNIT),
    "missing_testcases": ("missing", TEST_CASE),
    "missing_testsuites": ("missing", XUNIT),
    "updated_testcases": ("update", TEST_CASE),
}



# encoding: utf-8
# pylint: disable=missing-docstring,no-self-use

from __future__ import unicode_literals

import copy
import io
import logging
import os
from collections import OrderedDict

import pytest
import yaml

from tests import conf


# pylint: disable=invalid-name,too-many-ancestors
def ordered_load(stream, Loader=yaml.Loader):  # noqa: N803
    class OrderedLoader(Loader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return OrderedDict(loader.construct_pairs(node))

    OrderedLoader.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping)

    return yaml.load(stream, OrderedLoader)


def get_config():
    conf_file = os.path.join(conf.DATA_PATH, "polarion_tools.yaml")
    with io.open(conf_file, encoding="utf-8") as input_file:
        return ordered_load(input_file)


_config = get_config()


@pytest.fixture(scope="function")
def config():
    return copy.deepcopy(_config)


@pytest.fixture(scope="module")
def config_modulescope():
    return copy.deepcopy(_config)


class SimpleFormatter(object):
    def format(self, record):
        message = record.getMessage()
        if isinstance(message, bytes):
            message = message.decode("utf-8")
        return message


@pytest.yield_fixture
def captured_log():
    buff = io.StringIO()
    handler = logging.StreamHandler(buff)
    handler.setFormatter(SimpleFormatter())

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)

    yield buff

    logger.handlers.remove(handler)



# -*- coding: utf-8 -*-
"""
Configuration loading.
"""

from __future__ import absolute_import, unicode_literals

# pylint: disable=unused-import
from dump2polarion.configuration import get_config  # noqa



# -*- coding: utf-8 -*-
"""
Create missing requirements and create list of requirements names and IDs.
"""

from __future__ import absolute_import, unicode_literals

import io
import json
import logging

from cfme_testcases import cfme_parsereq
from cfme_testcases.exceptions import TestcasesException
from dump2polarion import utils as d2p_utils
from dump2polarion.exporters import requirements_exporter

# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


class RequirementsXML(object):
    """Creates requirements XML."""

    def __init__(self, config, testcases_json, requirements_data=None, transform_func=None):
        self.config = config
        self.testcases_json = testcases_json
        self.transform_func = transform_func
        self._requirements_data = requirements_data

    def get_requirements_from_testcases(self):
        """Gets requirements used in test cases."""
        with io.open(self.testcases_json, encoding="utf-8") as input_json:
            testcases = json.load(input_json)["testcases"]

        requirements = set()
        for testcase in testcases:
            linked_items = testcase.get("linked-items")
            if linked_items:
                requirements.update(linked_items)

        requirements_data = [{"title": req} for req in requirements]
        return requirements_data

    @property
    def requirements_data(self):
        """Gets requirements data."""
        if self._requirements_data:
            return self._requirements_data

        try:
            self._requirements_data = cfme_parsereq.get_requirements()
        except TestcasesException:
            self._requirements_data = self.get_requirements_from_testcases()
        return self._requirements_data

    def gen_requirements(self):
        """Generates the requirements XML string using requirements data."""
        return gen_requirements_xml_str(self.requirements_data, self.config, self.transform_func)

    def get_requirements_xml_root(self):
        """Gets the requirements XML root."""
        req_xml_str = self.gen_requirements()
        return d2p_utils.get_xml_root_from_str(req_xml_str)


def gen_requirements_xml_str(requirements_data, config, transform_func=None):
    """Generates the requirements XML string."""
    requirements = requirements_exporter.RequirementExport(
        requirements_data, config, transform_func
    )
    return requirements.export()


def get_requirements_xml_root(config, testcases_json, requirements_data=None, transform_func=None):
    """Gets the requirements XML root."""
    return RequirementsXML(
        config, testcases_json, requirements_data=requirements_data, transform_func=transform_func
    ).get_requirements_xml_root()



# -*- coding: utf-8 -*-
"""
Exceptions.
"""

# pylint: disable=unused-import
from dump2polarion.exceptions import Dump2PolarionException, NothingToDoException  # noqa


class TestcasesException(Dump2PolarionException):
    """Testcases exception."""



# -*- coding: utf-8 -*-
"""
Methods for generating and working with testsuites XML files.
"""

from __future__ import absolute_import, unicode_literals

import datetime
import logging

from cfme_testcases import utils
from cfme_testcases.exceptions import TestcasesException
from dump2polarion import utils as d2p_utils
from dump2polarion.exporters import xunit_exporter

# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


def gen_testsuites_xml_str(config, testrun_id, tests_data_json, transform_func=None):
    """Generates the testsuites XML string."""
    assert testrun_id
    try:
        results = utils.load_json_file(tests_data_json)["results"]
    except Exception as err:
        raise TestcasesException("Cannot load results from `{}`: {}".format(tests_data_json, err))
    testsuites_data = xunit_exporter.ImportedData(results, testrun_id)
    testsuites = xunit_exporter.XunitExport(testrun_id, testsuites_data, config, transform_func)
    return testsuites.export()


def get_testsuites_xml_root(config, testrun_id, tests_data_json, transform_func=None):
    """Returns content of testsuite XML file for xunit importer."""
    testrun_id = (
        testrun_id
        or d2p_utils.get_testrun_id_config(config)
        or "IMPORT_{:%Y%m%d%H%M%S}".format(datetime.datetime.now())
    )
    testsuites_str = gen_testsuites_xml_str(
        config, testrun_id, tests_data_json, transform_func=transform_func
    )
    return d2p_utils.get_xml_root_from_str(testsuites_str)



# -*- coding: utf-8 -*-
"""
Filter missing testcases and testcases for update.
"""

from __future__ import absolute_import, unicode_literals

import copy

from cfme_testcases.exceptions import TestcasesException
from dump2polarion import properties


class XMLFilters(object):
    """Filters for testcases in Importer's XML files."""

    def __init__(self, testcases_root, testsuites_root, missing, data_in_code=False):
        self.testcases_root = testcases_root
        self.testsuites_root = testsuites_root
        self.missing = missing or set()
        self.data_in_code = data_in_code

    def get_missing_testcases(self):
        """Gets testcases missing in Polarion."""
        xml_root = copy.deepcopy(self.testcases_root)

        testcase_instances = xml_root.findall("testcase")
        # Expect that in ID is the value we want.
        # In case of "lookup-method: name" it's test case title.
        attr = "id"

        for testcase in testcase_instances:
            tc_id = testcase.get(attr)
            if tc_id and tc_id not in self.missing:
                xml_root.remove(testcase)

        if not xml_root.findall("testcase"):
            return None

        return xml_root

    def get_missing_testsuites(self):
        """Gets testcases missing in testrun."""
        xml_root = copy.deepcopy(self.testsuites_root)

        # don't want to use/update any of these properties when adding tests to testrun
        for prop in ("polarion-group-id", "polarion-testrun-template-id", "polarion-testrun-title"):
            properties.remove_property(xml_root, prop)

        testsuite = xml_root.find("testsuite")
        testcase_parent = testsuite
        testcase_instances = testcase_parent.findall("testcase")
        attr = "name"

        for testcase in testcase_instances:
            # try to get test case ID first and if it fails, get name
            try:
                tc_id_prop = testcase.xpath('.//property[@name = "polarion-testcase-id"]')[0]
                tc_id = tc_id_prop.get("value")
            except IndexError:
                tc_id = testcase.get(attr)
            if tc_id and tc_id not in self.missing:
                testcase_parent.remove(testcase)

        if not testcase_parent.findall("testcase"):
            return None

        testcase_parent.set("tests", str(len(testcase_parent.findall("testcase"))))
        testcase_parent.attrib.pop("errors", None)
        testcase_parent.attrib.pop("failures", None)
        testcase_parent.attrib.pop("skipped", None)

        return xml_root

    def get_updated_testcases(self):
        """Gets testcases that will be updated in Polarion."""
        xml_root = copy.deepcopy(self.testcases_root)

        testcase_instances = xml_root.findall("testcase")
        attr = "id"

        for testcase in testcase_instances:
            tc_id = testcase.get(attr)
            if tc_id is not None and tc_id in self.missing:
                xml_root.remove(testcase)
                continue

            if self.data_in_code:
                continue

            # source not authoritative, don't update custom-fields
            cfields_parent = testcase.find("custom-fields")
            cfields_instances = cfields_parent.findall("custom-field")
            for field in cfields_instances:
                field_id = field.get("id")
                if field_id not in ("automation_script", "caseautomation"):
                    cfields_parent.remove(field)

        if not xml_root.findall("testcase"):
            return None

        return xml_root


def check_xml_roots(testcases_root, testsuites_root):
    """Checks that the XML files are in expected format."""
    if testcases_root is not None and testcases_root.tag != "testcases":
        raise TestcasesException("XML file is not in expected format.")
    if testsuites_root is not None and testsuites_root.tag != "testsuites":
        raise TestcasesException("XML file is not in expected format.")


def add_filtered_xmls(xmls_container, missing, data_in_code=False):
    """Returns modified XMLs with testcases and testsuites."""
    testcases_root, testsuites_root = xmls_container.testcases, xmls_container.testsuites
    check_xml_roots(testcases_root, testsuites_root)

    xml_filters = XMLFilters(testcases_root, testsuites_root, missing, data_in_code=data_in_code)

    if missing:
        if testcases_root is not None:
            xmls_container.missing_testcases = xml_filters.get_missing_testcases()
        if testsuites_root is not None:
            xmls_container.missing_testsuites = xml_filters.get_missing_testsuites()

    if testcases_root is not None:
        xmls_container.updated_testcases = xml_filters.get_updated_testcases()



# -*- coding: utf-8 -*-
"""
Collects all the XML files needed for import.
"""

from __future__ import absolute_import, unicode_literals

import copy
import logging

from cfme_testcases import submit, utils
from cfme_testcases.collector import filters, missing
from cfme_testcases.exceptions import NothingToDoException, TestcasesException
from dump2polarion import properties

# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


class XMLsCollector(object):
    """Methods for collecting XML files for import."""

    # pylint: disable=too-many-arguments
    def __init__(self, args, submit_args, config, testsuites_root, testcases_root):
        self.args = args
        self.submit_args = submit_args
        self.config = config
        self.testsuites_root = testsuites_root
        self.testcases_root = testcases_root

    def _initial_submit(self, xmls_container):
        """Submits XML to Polarion and saves the log file returned by the message bus."""
        if self.args.use_svn and not self.args.testrun_init:
            # no need to submit, SVN is used to generate list of missing testcases
            return None

        if not self.submit_args:
            raise NothingToDoException(
                "Instructed not to submit and as the import log is missing, "
                "there's nothing more to do."
            )

        if self.testsuites_root is None:
            raise TestcasesException("Cannot init testrun, testsuites XML not generated.")

        xml_root = copy.deepcopy(self.testsuites_root)

        if self.args.testrun_init:
            # want to init new test run
            dry_run = self.submit_args.get("dry_run") or False
            if self.args.testrun_title:
                properties.xunit_fill_testrun_title(xml_root, self.args.testrun_title)
        else:
            # want to just get the log file without changing anything
            dry_run = True
            # don't want to use template with dry-run
            properties.remove_property(xml_root, "polarion-testrun-template-id")

        init_sargs = self.submit_args.copy()
        init_sargs["dry_run"] = dry_run

        xmls_container.init_testsuites = xml_root
        return submit.initial_submit(init_sargs, self.config, xmls_container, self.args.output_dir)

    def get_multiple(self):
        """Gets filtered XMLs separated into multiple files."""
        xmls_container = utils.XMLsContainer()
        xmls_container.testcases = self.testcases_root
        xmls_container.testsuites = self.testsuites_root if self.args.testrun_id else None

        init_logname = self._initial_submit(xmls_container)
        missing_testcases = missing.get_missing_in_polarion(
            self.config, self.testcases_root, init_logname, self.args.use_svn
        )

        filters.add_filtered_xmls(
            xmls_container, missing_testcases, data_in_code=self.args.data_in_code
        )

        if self.args.no_testcases_update:
            xmls_container.updated_testcases = None

        # we don't want to submit unfiltered files
        xmls_container.testcases = None
        xmls_container.testsuites = None

        return xmls_container

    def get_single(self):
        """Gets single XML for each importer."""
        xmls_container = utils.XMLsContainer()
        xmls_container.testcases = self.testcases_root

        if self.args.mark_missing:
            # add data for testcases missing in source code
            marked_xml_root = missing.mark_missing_in_source(
                self.config, self.testcases_root, self.args.use_svn
            )
            xmls_container.testcases = marked_xml_root

        if not self.args.testrun_id:
            logger.warning("Not updating testrun, testrun ID not specified.")
            return xmls_container

        if self.args.testrun_init:
            if self.args.testrun_title:
                properties.xunit_fill_testrun_title(self.testsuites_root, self.args.testrun_title)
            xmls_container.testsuites = self.testsuites_root
            return xmls_container

        testcases_logname = None
        if self.submit_args:
            # Not initializing new testrun, we need to add just new testcases to existing testrun.
            # We'll submit testcases and parse produced log file to figure out what testcases
            # are missing in the testrun.
            testcases_logname = submit.import_testcases(
                self.submit_args, self.config, xmls_container, self.args.output_dir
            )

        if not (testcases_logname or self.args.use_svn):
            raise NothingToDoException("The import log is missing, there's nothing more to do.")

        missing_testcases = missing.get_missing_in_polarion(
            self.config, self.testcases_root, testcases_logname, self.args.use_svn
        )

        missing_testsuites_xml_root = filters.XMLFilters(
            None, self.testsuites_root, missing_testcases
        ).get_missing_testsuites()

        xmls_container.missing_testsuites = missing_testsuites_xml_root

        return xmls_container


def get_xmls(args, submit_args, config, testsuites_root, testcases_root):
    """Returns XMLsContainer object with all relevant XML files."""
    xmls_collector = XMLsCollector(args, submit_args, config, testsuites_root, testcases_root)
    if args.no_testcases_update or not args.data_in_code:
        return xmls_collector.get_multiple()
    return xmls_collector.get_single()



# -*- coding: utf-8 -*-
"""
Find testcases that are missing in Polarion or source code.
"""

from __future__ import absolute_import, unicode_literals

import copy
import logging
import os

from lxml import etree

from cfme_testcases import testcases, utils
from cfme_testcases.collector import testcases_svn
from cfme_testcases.exceptions import TestcasesException
from dump2polarion import parselogs

# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


class MissingInPolarion(object):
    """Testcases missing in Polarion."""

    def __init__(self, config, all_testcases, init_logname, svn_dir):
        self.config = config
        self.all_testcases = all_testcases
        self.init_logname = init_logname or ""
        self.svn_dir = svn_dir

    def get_missing_from_log(self):
        """Gets missing testcases from log file."""
        lookup_method = self.config.get("xunit_import_properties") or {}
        lookup_method = lookup_method.get("polarion-lookup-method") or "id"
        parsed_log = parselogs.parse(self.init_logname)

        if lookup_method == "name":
            missing = [item.name for item in parsed_log.new_items]
        elif lookup_method == "custom":
            missing = [item.custom_id for item in parsed_log.new_items]
        elif lookup_method == "id":
            missing = [item.id for item in parsed_log.new_items]

        return set(missing)

    def get_missing_from_svn(self):
        """Gets missing testcases using SVN repo."""
        lookup_by = testcases.get_lookup_by(self.config)
        return testcases_svn.get_missing_in_polarion(
            self.svn_dir, self.all_testcases, lookup_by=lookup_by
        )

    def get_missing(self):
        """Returns list of missing testcases."""
        utils.check_lookup_methods(self.config)
        log_exists = os.path.isfile(self.init_logname)
        if self.svn_dir and not log_exists:
            missing = self.get_missing_from_svn()
        elif not log_exists:
            raise TestcasesException(
                "The submit log file `{}` doesn't exist.".format(self.init_logname)
            )
        else:
            missing = self.get_missing_from_log()
        return missing


def get_missing_in_polarion(config, testcases_root, init_logname, svn_dir):
    """Returns list of missing testcases in Polarion."""
    all_testcases = testcases.get_all_testcases(testcases_root)
    return MissingInPolarion(config, all_testcases, init_logname, svn_dir).get_missing()


def _get_missing_in_source(config, testcases_root, svn_dir):
    """Gets set of testcases missing in source code."""
    all_testcases = testcases.get_all_testcases(testcases_root)
    lookup_by = testcases.get_lookup_by(config)
    return testcases_svn.get_missing_in_source(svn_dir, all_testcases, lookup_by=lookup_by)


def _append_missing_testcases(parent_element, missing_ids):
    """Appends testcases data to XML for testcases missing in source code."""
    for missing_id in missing_ids:
        testcase = etree.SubElement(parent_element, "testcase", {"id": missing_id})
        custom_fields_el = etree.SubElement(testcase, "custom-fields")
        etree.SubElement(
            custom_fields_el, "custom-field", {"id": "tags", "content": "NOT_FOUND_IN_SOURCES"}
        )


def mark_missing_in_source(config, testcases_root, svn_dir):
    """Mark testcases missing in source code."""
    if not svn_dir:
        logger.warning("Not marking testcases missing in source code, SVN repo not specified.")
        return testcases_root

    utils.check_lookup_methods(config)
    xml_root = copy.deepcopy(testcases_root)
    missing_ids = _get_missing_in_source(config, xml_root, svn_dir)
    _append_missing_testcases(xml_root, missing_ids)
    return xml_root



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring,no-self-use,too-few-public-methods,protected-access


from __future__ import unicode_literals

import copy
import io
import os
import re

from cfme_testcases import testcases, testsuites, utils
from cfme_testcases.collector import filters, missing
from dump2polarion import utils as d2p_utils
from dump2polarion.exporters import transform
from tests import conf

TESTS_JSON = os.path.join(conf.DATA_PATH, "tests_data.json")
INIT_LOG = os.path.join(conf.DATA_PATH, "job-init.log")


def get_testcases_transform(config, plugin_name):
    """Return test cases transformation function for Insights."""

    parametrize = config.get("insights_parametrize", True)
    testcase_append = plugin_name
    seen_ids = set()

    def testcase_transform(testcase):
        """Test cases transform for Insights."""
        testcase = copy.deepcopy(testcase)

        transform.setup_parametrization(testcase, parametrize)

        testcase_id = transform.get_testcase_id(testcase, testcase_append)
        testcase["id"] = testcase_id

        if parametrize and testcase_id in seen_ids:
            return None
        seen_ids.add(testcase_id)

        transform.parse_rst_description(testcase)
        transform.add_unique_runid(testcase, "id123")
        transform.add_automation_link(testcase)

        return testcase

    return testcase_transform


def get_xunit_transform(config, plugin_name):
    """Return result transformation function for Insights."""
    skip_searches = ["SKIPME:", "Skipping due to these blockers", "BZ ?[0-9]+", "GH ?#?[0-9]+"]
    skips = re.compile("(" + ")|(".join(skip_searches) + ")")

    parametrize = config.get("insights_parametrize", True)
    testcase_append = plugin_name

    def results_transform(result):
        """Results transform for Insights."""
        verdict = result.get("verdict")
        if not verdict:
            return None

        result = copy.deepcopy(result)

        transform.setup_parametrization(result, parametrize)
        transform.include_class_in_title(result)
        transform.insert_source_info(result)

        result["id"] = transform.get_testcase_id(result, testcase_append)

        verdict = verdict.strip().lower()
        # we want to submit PASS and WAIT results
        if verdict in transform.Verdicts.PASS + transform.Verdicts.WAIT:
            return result
        comment = result.get("comment")
        # ... and SKIP results where there is a good reason (blocker etc.)
        if verdict in transform.Verdicts.SKIP and comment and skips.search(comment):
            # found reason for skip
            result["comment"] = comment.replace("SKIPME: ", "").replace("SKIPME", "")
            return result
        if verdict in transform.Verdicts.FAIL and comment and "FAILME" in comment:
            result["comment"] = comment.replace("FAILME: ", "").replace("FAILME", "")
            return result
        # we don't want to report this result if here
        return None

    return results_transform


def test_filtered(config):
    fname = "testcases_updated.xml"

    testsuites_root = testsuites.get_testsuites_xml_root(
        config, "IMPORT123", TESTS_JSON, transform_func=get_xunit_transform(config, "vulnerability")
    )
    testcases_root = testcases.get_testcases_xml_root(
        config,
        TESTS_JSON,
        set_requirements=False,
        transform_func=get_testcases_transform(config, "vulnerability"),
    )

    missing_testcases = missing.get_missing_in_polarion(config, testcases_root, INIT_LOG, None)
    xmls_container = utils.XMLsContainer()
    xmls_container.testcases = testcases_root
    xmls_container.testsuites = testsuites_root
    filters.add_filtered_xmls(xmls_container, missing_testcases, data_in_code=True)
    # pylint: disable=no-member
    updated_testcases = d2p_utils.prettify_xml(xmls_container.updated_testcases)

    with io.open(os.path.join(conf.DATA_PATH, fname), encoding="utf-8") as input_xml:
        parsed = input_xml.read()
    assert parsed == updated_testcases



# -*- coding: utf-8 -*-
"""
Testcases data from Polarion SVN repo.
"""

from __future__ import absolute_import, unicode_literals

import logging
import os
import re
from collections import defaultdict

from cfme_testcases.exceptions import TestcasesException
from dump2polarion import svn_polarion

# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


class PolarionTestcases(object):
    """Loads and access Polarion testcases."""

    TEST_PARAM = re.compile(r"\[.*\]")
    tc_cache = defaultdict(dict)

    def __init__(self, repo_dir):
        self.repo_dir = repo_dir
        self.available_testcases = {}
        wi_cache = self.tc_cache[repo_dir].get("wi_cache")
        if wi_cache:
            self.wi_cache = wi_cache
        else:
            self.wi_cache = svn_polarion.WorkItemCache(repo_dir)
            self.tc_cache[repo_dir]["wi_cache"] = self.wi_cache

    def load_active_testcases(self, lookup_by="title"):
        """Creates dict of all active testcases names and ids."""
        cached_lookup_by = self.tc_cache[self.repo_dir].get(lookup_by)
        if cached_lookup_by:
            self.available_testcases = cached_lookup_by
            return

        cases = {}
        for item in self.wi_cache.get_all_items():
            if item.get("type") != "testcase":
                continue
            case_status = item.get("status")
            if not case_status or case_status == "inactive":
                continue

            case_title = item.get(lookup_by)
            case_id = item.get("work_item_id")
            try:
                cases[case_title].append(case_id)
            except KeyError:
                cases[case_title] = [case_id]

        self.available_testcases = cases
        self.tc_cache[self.repo_dir][lookup_by] = cases

    def __getitem__(self, item):
        return self.available_testcases[item]

    def __iter__(self):
        return iter(self.available_testcases)

    def __len__(self):
        return len(self.available_testcases)

    def __contains__(self, item):
        return item in self.available_testcases

    def __repr__(self):
        return "<Testcases {}>".format(self.available_testcases)


def _load_testcases(repo_dir, lookup_by):
    polarion_testcases = PolarionTestcases(os.path.expanduser(repo_dir))
    try:
        polarion_testcases.load_active_testcases(lookup_by=lookup_by)
    except Exception as err:
        raise TestcasesException(
            "Failed to load testcases from SVN repo `{}`: {}".format(repo_dir, err)
        )
    if not polarion_testcases:
        raise TestcasesException("No testcases loaded from SVN repo `{}`.".format(repo_dir))

    return polarion_testcases


def get_missing_in_polarion(repo_dir, testcase_identifiers, lookup_by="title"):
    """Gets set of testcases missing in Polarion."""
    polarion_testcases = _load_testcases(repo_dir, lookup_by)
    return set(testcase_identifiers) - set(polarion_testcases)


def get_missing_in_source(repo_dir, testcase_identifiers, lookup_by="title"):
    """Gets set of testcases missing in source code but present in Polarion."""
    polarion_testcases = _load_testcases(repo_dir, lookup_by)
    return set(polarion_testcases) - set(testcase_identifiers)



# -*- coding: utf-8 -*-
"""
Methods for generating and working with testcases XML files.
"""

from __future__ import absolute_import, unicode_literals

import logging

import six

from cfme_testcases import utils
from cfme_testcases.exceptions import TestcasesException
from dump2polarion import utils as d2p_utils
from dump2polarion.exporters import testcases_exporter

# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


class TestcasesXML(object):
    """Generate XML for Testcases Importer."""

    requirements_lookup = "name"

    def __init__(self, config, tests_data_json, set_requirements=True, transform_func=None):
        self.config = config
        self.tests_data_json = tests_data_json
        self.set_requirements = set_requirements
        self.transform_func = transform_func

    def set_requirements_lookup(self, testcase_rec):
        """Sets lookup methods for requirements in testcase."""
        if testcase_rec.get("linked-items-lookup-method") in ("id", "name"):
            return
        testcase_rec["linked-items-lookup-method"] = self.requirements_lookup

    def setup_requirements(self, testcases_data):
        """Configures requirements in testcase."""
        for testcase_rec in testcases_data:
            requirement_names = testcase_rec.get("linked-items")
            if not requirement_names:
                continue
            if not self.set_requirements:
                testcase_rec.pop("linked-items", None)
                continue
            if isinstance(requirement_names, (dict, six.string_types)):
                requirement_names = [requirement_names]
            self.set_requirements_lookup(testcase_rec)

    def gen_testcases_xml_str(self):
        """Generates the testcases XML string."""
        try:
            testcases_data = utils.load_json_file(self.tests_data_json)["testcases"]
        except Exception as err:
            raise TestcasesException(
                "Cannot load test cases from `{}`: {}".format(self.tests_data_json, err)
            )
        self.setup_requirements(testcases_data)
        testcases = testcases_exporter.TestcaseExport(
            testcases_data, self.config, self.transform_func
        )
        return testcases.export()


def get_testcases_xml_root(config, tests_data_json, set_requirements=True, transform_func=None):
    """Returns content of testcase XML file for testcase importer."""
    testcases_xml = TestcasesXML(
        config, tests_data_json, set_requirements=set_requirements, transform_func=transform_func
    )
    testcases_str = testcases_xml.gen_testcases_xml_str()
    return d2p_utils.get_xml_root_from_str(testcases_str)


def get_all_testcases(testcases_root):
    """Gets all testcases from XML."""
    if testcases_root.tag != "testcases":
        raise TestcasesException("XML file is not in expected format.")

    testcase_instances = testcases_root.findall("testcase")
    # Expect that in ID is the value we want.
    # In case of "lookup-method: name" it's test case title.
    attr = "id"

    for testcase in testcase_instances:
        tc_id = testcase.get(attr)
        if tc_id:
            yield tc_id


def get_lookup_by(config):
    """Gets lookup-method."""
    import_props = config.get("testcase_import_properties") or {}
    lookup_by = import_props.get("lookup-method") or "id"
    if lookup_by == "name":
        lookup_by = "title"
    elif lookup_by == "custom":
        lookup_by = import_props.get("polarion-custom-lookup-method-field-id") or "testCaseID"
    elif lookup_by == "id":
        lookup_by = "work_item_id"
    return lookup_by



# -*- coding: utf-8 -*-
"""
Gets CFME requirements data out of `REG_PATH` file. Used just in CFME, ignored if not present.
"""

from __future__ import absolute_import, unicode_literals

import ast

from cfme_testcases.exceptions import TestcasesException

REQ_PATH = "cfme/test_requirements.py"


def _get_tree(filename):
    try:
        with open(filename) as infile:
            source = infile.read()

        tree = ast.parse(source, filename=filename)
    except Exception as err:
        raise TestcasesException("Cannot parse `{}`: {}".format(filename, err))

    return tree


def parse_requirements_file(tree, filename):
    """Parse the requirements info out of REQ_PATH file."""
    if not tree:
        tree = _get_tree(filename)

    requirements_data = []
    for node in tree.body:
        if not isinstance(node, ast.Assign):
            continue

        try:
            func = node.value.func
            req_name = node.value.args[0].s
            if not (
                req_name
                and func.value.value.id == "pytest"
                and func.value.attr == "mark"
                and func.attr == "requirement"
            ):
                continue
        except AttributeError:
            continue

        requirement_record = {"title": req_name}

        try:
            keywords = node.value.keywords
        except AttributeError:
            keywords = {}

        for keyword in keywords:
            requirement_record[keyword.arg] = keyword.value.s

        requirements_data.append(requirement_record)

    return requirements_data


def get_requirements():
    """Gets the requirements info."""
    return parse_requirements_file(None, REQ_PATH)



# -*- coding: utf-8 -*-
"""
Run pytest --collect-only and generate JSONs.
"""

from __future__ import absolute_import, unicode_literals

import logging
import os
import subprocess
import sys

from cfme_testcases import consts
from cfme_testcases.exceptions import TestcasesException

# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


def json_ready(tests_data_json):
    """Check if the needed JSON file is already available."""
    if tests_data_json and not os.path.exists(tests_data_json):
        raise TestcasesException(
            "The testcases JSON file `{}` doesn't exist.".format(tests_data_json)
        )
    if not tests_data_json:
        return False
    return True


def collect(pytest_collect, tests_data_json):
    """Collects testcases data."""
    retval = None

    if not json_ready(tests_data_json):
        tests_data_json = consts.TESTS_DATA_JSON
        retval = run_pytest(pytest_collect)

    return retval, tests_data_json


def check_json():
    """Check that the JSON files were generated."""
    if not os.path.exists(consts.TESTS_DATA_JSON):
        raise TestcasesException("The JSON file `{}` doesn't exist.".format(consts.TESTS_DATA_JSON))


def check_environment():
    """Checks the environment for running pytest."""
    if not os.path.exists(".git"):
        raise TestcasesException("Not launched from the top-level directory.")
    # check that running in virtualenv
    if not hasattr(sys, "real_prefix"):
        raise TestcasesException("Not running in virtual environment.")


def cleanup():
    """Deletes JSON file generated during previous runs."""
    try:
        os.remove(consts.TESTS_DATA_JSON)
    except OSError:
        pass


def run_pytest(pytest_collect):
    """Runs the pytest command."""
    if not pytest_collect:
        raise TestcasesException(
            "The `pytest_collect` command for collecting testcases was not specified."
        )

    pytest_retval = 250
    check_environment()
    cleanup()

    pytest_args = pytest_collect.split(" ")

    logger.info("Generating the JSON using '%s'", pytest_collect)
    with open(os.devnull, "w") as devnull:
        pytest_proc = subprocess.Popen(pytest_args, stdout=devnull, stderr=devnull)
        try:
            pytest_retval = pytest_proc.wait()
        # pylint: disable=broad-except
        except Exception:
            try:
                pytest_proc.terminate()
            except OSError:
                pass
            pytest_proc.wait()

    check_json()
    return pytest_retval



# -*- coding: utf-8 -*-
"""
Main CLI.
"""

from __future__ import absolute_import, unicode_literals

import argparse
import logging

from box import Box

from cfme_testcases import pytest_collect, requirements, submit, testcases, testsuites, utils
from cfme_testcases.collector import xmls_collector
from cfme_testcases.exceptions import Dump2PolarionException, NothingToDoException
from dump2polarion import configuration
from dump2polarion import utils as d2p_utils

# pylint: disable=invalid-name
logger = logging.getLogger(__name__)


def get_args(args=None):
    """Get command line arguments."""
    parser = argparse.ArgumentParser(description="cfme-testcases")
    parser.add_argument("-t", "--testrun-id", help="Polarion test run id")
    parser.add_argument("-o", "--output_dir", help="Directory for saving generated XML files")
    parser.add_argument(
        "-n", "--no-submit", action="store_true", help="Don't submit generated XML files"
    )
    parser.add_argument(
        "--testrun-init", action="store_true", help="Create and initialize new testrun"
    )
    parser.add_argument("--testrun-title", help="Title to set for the new testrun")
    parser.add_argument(
        "--data-in-code",
        action="store_true",
        help="Source code is an authoritative source of data.",
    )
    parser.add_argument("--user", help="Username to use to submit to Polarion")
    parser.add_argument("--password", help="Password to use to submit to Polarion")
    parser.add_argument("--tests-data", help="Path to JSON file with tests data")
    parser.add_argument("--config", help="Path to polarion_tools config YAML")
    parser.add_argument("--dry-run", action="store_true", help="Dry run, don't update anything")
    parser.add_argument("--no-requirements", action="store_true", help="Don't import requirements")
    parser.add_argument(
        "--no-testcases-update", action="store_true", help="Don't update existing testcases"
    )
    parser.add_argument("--no-verify", action="store_true", help="Don't verify import success")
    parser.add_argument(
        "--verify-timeout",
        type=int,
        default=900,
        metavar="SEC",
        help="How long to wait (in seconds) for verification of submission success"
        " (default: %(default)s)",
    )
    parser.add_argument(
        "--use-svn", metavar="SVN_REPO", help="Path to SVN repo with Polarion project"
    )
    parser.add_argument(
        "--mark-missing", action="store_true", help="Mark testcases missing in source code"
    )
    parser.add_argument("--log-level", help="Set logging to specified level")
    parsed = parser.parse_args(args)

    if parsed.mark_missing and not (parsed.use_svn and parsed.data_in_code):
        parser.error("The --mark-missing argument requires --use-svn and --data-in-code")
    if parsed.testrun_title and not parsed.testrun_init:
        parser.error("The --testrun-title argument requires --testrun-init")

    return parsed


def process_args(args):
    """Processes passed arguments."""
    passed_args = args
    if isinstance(args, argparse.Namespace):
        passed_args = vars(passed_args)
    elif hasattr(args, "to_dict"):
        passed_args = passed_args.to_dict()

    return Box(passed_args, frozen_box=True, default_box=True)


# pylint: disable=too-many-locals,too-many-arguments
def update_testcases(
    args,
    config,
    requirements_data=None,
    requirements_transform_func=None,
    xunit_transform_func=None,
    testcases_transform_func=None,
):
    """Testcases update main function."""
    assert isinstance(requirements_data, list) if requirements_data is not None else True

    args = process_args(args)
    submit_args = utils.get_submit_args(args) if not args.no_submit else None

    try:
        __, tests_data_json = pytest_collect.collect(config.get("pytest_collect"), args.tests_data)

        requirements_root = None

        if not args.no_requirements:
            requirements_root = requirements.get_requirements_xml_root(
                config,
                tests_data_json,
                requirements_data=requirements_data,
                transform_func=requirements_transform_func,
            )

        testsuites_root = testsuites.get_testsuites_xml_root(
            config, args.testrun_id, tests_data_json, transform_func=xunit_transform_func
        )
        testcases_root = testcases.get_testcases_xml_root(
            config,
            tests_data_json,
            set_requirements=requirements_root is not None,
            transform_func=testcases_transform_func,
        )

        xmls_container = xmls_collector.get_xmls(
            args, submit_args, config, testsuites_root, testcases_root
        )
        xmls_container.requirements = requirements_root

        if args.no_submit or args.output_dir:
            utils.save_xmls(args.output_dir, xmls_container)

        submit.submit_all(submit_args, config, xmls_container, args.output_dir)
    except NothingToDoException as einfo:
        logger.info(einfo)
        return 0
    except Dump2PolarionException as err:
        logger.fatal(err)
        return 1
    return 0


def main(
    args=None,
    requirements_transform_func=None,
    xunit_transform_func=None,
    testcases_transform_func=None,
):
    """Main function for CLI."""
    args = get_args(args)
    d2p_utils.init_log(args.log_level)

    try:
        config = configuration.get_config(args.config)
    except Dump2PolarionException as err:
        logger.fatal(err)
        return 1

    return update_testcases(
        args,
        config,
        requirements_transform_func=requirements_transform_func,
        xunit_transform_func=xunit_transform_func,
        testcases_transform_func=testcases_transform_func,
    )
